'''
Created on 17 Dec 2023

@author: jacklok
'''

from datetime import datetime, date
from trexmodel import program_conf
from dateutil.relativedelta import relativedelta
import logging
from trexprogram.reward_program.reward_program_base import EntitledVoucherSummary
from trexmodel.models.datastore.voucher_models import MerchantVoucher
from trexmodel.models.datastore.reward_models import CustomerEntitledVoucher
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.utils.model.model_util import generate_transaction_id
from trexmodel.models.datastore.redeem_models import RedemptionCatalogueTransaction,\
    CustomerRedemption
from trexanalytics.bigquery_upstream_data_config import create_merchant_customer_redemption_upstream_for_merchant
from trexmodel.models.datastore.customer_model_helpers import update_customer_entiteld_voucher_summary_with_customer_new_voucher
from trexmodel.models.datastore.message_model_helper import create_redeem_catalogue_item_message
from trexprogram.utils.reward_program_helper import calculate_effective_date,\
    calculate_expiry_date
from trexconf.program_conf import REWARD_PROGRAM_DATE_FORMAT

logger = logging.getLogger('helper')

@model_transactional(desc='giveaway_redeem_catalogue_item')
def giveaway_redeem_catalogue_item(customer, redemption_catalogue, voucher_key):
    redeemed_datetime       = datetime.utcnow()
    redeem_reward_format    = redemption_catalogue.redeem_reward_format
    redeem_item_details     = next(filter(lambda x: x['voucher_key'] == voucher_key, redemption_catalogue.catalogue_settings.get('items')), None)
    redeem_reward_amount    = redeem_item_details.get('redeem_reward_amount')
    
    if redeem_reward_format in (program_conf.REWARD_FORMAT_POINT,program_conf.REWARD_FORMAT_STAMP) :
        reward_summary      = customer.reward_summary
        
        logger.debug('********************************')
        logger.debug('customer reward_summary=%s', reward_summary)
        logger.debug('********************************')
        
        if reward_summary.get(redeem_reward_format).get('amount') < redeem_reward_amount:
            raise Exception('Not sufficient reward amount to redeem')
    
        else:
            transaction_id = generate_transaction_id(prefix='r')
            redemption_catalogue_transction_summary = {
                                                        'redemption_catalogue_key'  : redemption_catalogue.key_in_str,
                                                        'voucher_key'               : voucher_key,
                                                        
                                                        }
            CustomerRedemption.create(customer, 
                                    reward_format                           = redeem_reward_format,
                                    redeemed_amount                         = redeem_reward_amount,
                                    redeemed_datetime                       = redeemed_datetime, 
                                    transaction_id                          = transaction_id,
                                    redemption_catalogue_transction_summary = redemption_catalogue_transction_summary,
                                    )
            
            redemption_catalogue_transaction = RedemptionCatalogueTransaction.create(redemption_catalogue,
                                                  voucher_key, 
                                                  customer,
                                                  transaction_id,
                                                  redeemed_datetime,
                                                  )
            
            reward_summary = __giveaway_voucher_from_redemption_catalogue_item(customer, redeem_item_details, transaction_id, redeemed_datetime)
            
            create_redeem_catalogue_item_message(customer, reward_summary.entitled_voucher_summary, redemption_catalogue_transaction)
            
            #if customer_redemption:
            #    logger.info('Going to create customer redemption upstream')
                #create_merchant_customer_redemption_upstream_for_merchant(customer_redemption, streamed_datetime=redeemed_datetime)
            
            return reward_summary    


def __giveaway_voucher_from_redemption_catalogue_item(customer, redeem_item_details, transaction_id, transact_datetime):
    logger.debug('---__giveaway_voucher_from_redemption_catalogue_item---')
    
    voucher_key             = redeem_item_details.get('voucher_key')
    effective_type          = redeem_item_details.get('effective_type')
    effective_value         = redeem_item_details.get('effective_value')
    effective_date_str      = redeem_item_details.get('effective_date')
    
    if effective_type == program_conf.REWARD_EFFECTIVE_TYPE_SPECIFIC_DATE:
        if is_not_empty(effective_date_str):
            effective_date = datetime.strptime(effective_date_str, REWARD_PROGRAM_DATE_FORMAT)
    else:
        effective_date = calculate_effective_date(effective_type, effective_value, start_date = transact_datetime)
    
    expiration_type         = redeem_item_details.get('expiration_type')
    expiration_value        = redeem_item_details.get('expiration_value')
     
    expiry_date             = calculate_expiry_date(expiration_type, expiration_value, start_date=effective_date)
    
    voucher_amount          = redeem_item_details.get('voucher_amount')
    
    merchant_voucher = MerchantVoucher.fetch(voucher_key)
    customer_entitled_voucher_list = []
    reward_summary = EntitledVoucherSummary()
    
    logger.debug('merchant_voucher=%s', merchant_voucher)
    logger.debug('voucher_amount=%s', voucher_amount)
    
    if merchant_voucher:
        entitled_voucher_summary = customer.entitled_voucher_summary
        for v in range(voucher_amount):
            customer_entitled_voucher = CustomerEntitledVoucher.create(
                                                            merchant_voucher,
                                                            customer, 
                                                            transaction_id      = transaction_id,
                                                            rewarded_datetime   = transact_datetime,
                                                            effective_date      = effective_date,
                                                            expiry_date         = expiry_date,
                                                            
                                                            )
            
            logger.debug('customer_entitled_voucher=%s', customer_entitled_voucher)
            
            update_customer_entiteld_voucher_summary_with_customer_new_voucher(entitled_voucher_summary, customer_entitled_voucher)
            customer_entitled_voucher_list.append(customer_entitled_voucher)
        
        customer.entitled_voucher_summary = entitled_voucher_summary    
        customer.put()
        
        reward_summary.add(merchant_voucher, 
                                   customer_entitled_voucher_list) 
        
    return reward_summary



from flask import Blueprint
import logging
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.admin_models import AppBannerFile
 
from flask.json import jsonify

app_api_bp = Blueprint('app_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/app')

logger = logging.getLogger('debug')


@app_api_bp.route('/ping', methods=['GET'])
def ping():
    return 'pong', 200

@app_api_bp.route('/app-package-info/app/<app_name>/platform/<platform>', methods=['GET'])
def app_package_info(app_name, platform):
    logger.debug('---app_package_info---')
    
    logger.info('app_name=%s, plaform=%s', app_name, platform)
    package_info = {}
    
    if platform=='android':
        if app_name == 'memberasia':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.0',
                        'can_update'    : False,
                        'update_url'    : '',    
        
                    }
        elif app_name == 'memberla':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.1',
                        'can_update'    : True,
                        'update_url'    : 'https://play.google.com/store/apps/details?id=com.augmigo.memberla',
        
                    }
        
    elif platform=='ios':
        
        if app_name == 'memberasia':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.0',
                        'can_update'    : False,
                        'update_url'    : '',
        
                    }
        elif app_name == 'memberla':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.1',
                        'can_update'    : False,
                        'update_url'    : '',
        
                    }
            
    elif platform=='window':
        
        if app_name == 'memberasia':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.0',
                        'can_update'    : False,
                        'update_url'    : '',
        
                    }
        elif app_name == 'memberla':
            package_info =  {
                        'app_name'      : app_name,
                        'platform'      : platform,
                        'version'       : '1.0.1',
                        'can_update'    : False,
                        'update_url'    : '',
        
                    }        
               
    return package_info

@app_api_bp.route('/settings', methods=['GET'])
def app_setting():
    logger.debug('---app_setting---')
    
    banner_file_list = []
    db_client = create_db_client(caller_info="app_setting")
    with db_client.context():
        
        result_listing = AppBannerFile.list()
        logger.debug('result_listing=%s', result_listing)
                
        if result_listing:
            for banner_file in result_listing:
                #banner_file_list.append(banner_file.to_dict(dict_properties=['banner_file_public_url','sequence'], show_key=False))
                banner_file_list.append({
                                        'image_url': banner_file.banner_file_public_url,
                                        'sequence': banner_file.sequence,
                                        })
                
    ##sorted_banner_file_list = sort_dict_list(banner_file_list, sort_attr_name='sequence')
    
    app_settings =  {
                        'banners': banner_file_list,
                        'message_box_settings'  : {
                                                'max_read_count'                : 2,
                                                'read_interval_in_second'       : 60,
                                                'list_message_pagination_limit' : 10,
                                                },
                        'account_settings'      :{
                                                'login_session_length_in_hour'  : 96,
                                                },
                        'outlet_order_enabled'  : False,
                        
        
                    }
    '''
    return create_rest_message(status_code=StatusCode.OK,
                                               **app_settings,
                                               )
    '''
    return app_settings

@app_api_bp.route('/promotions', methods=['GET'])
def list_app_promotions():
    logger.debug('---list_app_promotions---')
    result_listing = [
            
            {
                'title':'GoodTaste',
                'content':'10% Discount for member during June 2023',
            },
            {
                'title':'Anviet',
                'content':'Entitled RM10 cash voucher for miniumum RM50 spending per receipt during June 2023',
            }
            
        
        ]
    '''
    db_client = create_db_client(caller_info="list_app_promotions")
    with db_client.context():
        
        result_listing = AppPromotion.list()
        logger.debug('result_listing=%s', result_listing)
    '''
    
        
    return jsonify({
            'promotions': result_listing,
            })
    
@app_api_bp.route('/messages', methods=['GET'])
def list_app_messages():
    logger.debug('---list_app_messages---')
    result_listing = [
            {
                'title':'Welcome',
                'content':'Welcome on board',
            }
        ]
    '''
    db_client = create_db_client(caller_info="list_app_highlights")
    with db_client.context():
        
        result_listing = AppMessage.list()
        logger.debug('result_listing=%s', result_listing)
    '''
    
        
    return jsonify({
            'messages': result_listing,
            })    





from wtforms import StringField, DecimalField, validators
from trexadmin.forms.base_forms import ValidationBaseForm
from trexadmin.libs.wtforms import validators as custom_validator
from trexadmin.libs.wtforms.fields import OptionalDateTimeField, CurrencyField, JSONField
from flask_babel import gettext

class SalesTransactionForm(ValidationBaseForm):
    sales_amount                    = DecimalField('Sales Amount',[
                                                validators.InputRequired(message="Sales Amount is required"),
                                            ])
    
    tax_amount                      = DecimalField('Tax Amount',[
                                                validators.Optional()
                                            ])  
    
    invoice_id                      = StringField('Invoice No',[
                                                validators.Optional(),
                                                validators.Length(max=30, message="Invoice No length must not more than 30 characters")
                                            ])
    
    remarks                         = StringField('Remarks',[
                                                validators.Optional(),
                                                validators.Length(max=300, message="Remarks length must not more than 300 characters")
                                            ])
    
    transact_datetime               = OptionalDateTimeField('Transact Datetime', format='%d-%m-%Y %H:%M:%S') 
    
     
    
    




import random, logging
from flask import request
from flask.blueprints import Blueprint
from trexmodel.models.datastore.lucky_draw_models import LuckyDrawTicket
from trexmodel.utils.model.model_util import create_db_client
from flask.json import jsonify
from trexlib.utils.log_util import get_tracelog
from trexmodel.models.datastore.merchant_models import MerchantAcct
from trexapi.decorators.api_decorators import auth_token_required, user_auth_token_required
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.customer_models import Customer
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.transaction_models import CustomerTransaction
from datetime import datetime
from trexmodel import program_conf
from trexprogram.reward_program.lucky_draw_program import LuckyDrawRewardProgram
from trexapi.utils.api_helpers import create_api_message, StatusCode

lucky_draw_api_bp = Blueprint('lucky_draw_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/lucky-draw')

logger = logging.getLogger('debug')

@lucky_draw_api_bp.route('/draw', methods=['GET'])
def draw():

    # Define the prizes and their respective weights (i.e., the likelihood of winning each prize)
    prizes = {"First Prize": 1, "Second Prize": 4, "Third Prize": 10, "Consolation Prize": 40, "Try again": 144}
    
    # Get the total weight of all the prizes
    total_weight = sum(prizes.values())
    
    # Generate a random number between 1 and the total weight
    random_num = random.randint(1, total_weight)
    
    # Determine the prize based on the random number and the weights of the prizes
    draw_prize = None
    for prize, weight in prizes.items():
        if random_num <= weight:
            print(f"You've won the {prize}!")
            draw_prize = prize
            break
        else:
            random_num -= weight
            
    return draw_prize, 200

@lucky_draw_api_bp.route('/<ticket_key>/draw', methods=['POST'])
#@user_auth_token_required
def lucky_draw(ticket_key):
    
    drawed_data_in_json   = request.get_json()
    db_client = create_db_client(caller_info="lucky_draw")
    
    logger.debug('drawed_data_in_json=%s', drawed_data_in_json)
    
    selected_index = drawed_data_in_json.get('selected_index')
    
    logger.debug('selected_index=%s', selected_index)
    
    
    drawed_prize_sequence_indexes   = []
    
    try:
        with db_client.context():
            
            lucky_draw_ticket   = LuckyDrawTicket.get_by_ticket_key(ticket_key)
            logger.debug('lucky_draw_ticket=%s', lucky_draw_ticket)
        
        
        if lucky_draw_ticket:
            lucky_draw_ticket.patch_prize_image_base_url()
            drawed_prize_sequence_indexes = lucky_draw_ticket.drawed_details.get('drawed_prize_sequence_indexes')
            
            if lucky_draw_ticket.used==False:
                logger.debug('Going to draw')
                with db_client.context():
                    lucky_draw_ticket.draw(selected_index=selected_index)
                    
                drawed_prize_sequence_indexes = lucky_draw_ticket.drawed_details.get('drawed_prize_sequence_indexes')    
                    
                logger.debug('drawed_prize_sequence_indexes=%s', drawed_prize_sequence_indexes)
                
                return create_api_message(status_code=StatusCode.OK, 
                                           won_prize                        = lucky_draw_ticket.drawed_details.get('won_prize'),
                                           drawed_prize_sequence_indexes    = drawed_prize_sequence_indexes,
                                           selected_index                   = lucky_draw_ticket.drawed_details.get('selected_index'),
                                           drawed_datetime                  = lucky_draw_ticket.used_datetime.strftime('%d-%m-%Y %H:%M:%s'),
                                           )
            else:
                if drawed_prize_sequence_indexes is None:
                    with db_client.context(): 
                        lucky_draw_ticket.draw_update_prize_sequence(selected_index=selected_index)
                        drawed_prize_sequence_indexes = lucky_draw_ticket.drawed_details.get('drawed_prize_sequence_indexes')
                
                with db_client.context():
                    customer_acct = lucky_draw_ticket.customer_acct_entity
                    Customer.update_ticket_into_lucky_draw_ticket_summary(customer_acct, lucky_draw_ticket.to_configuration())        
                    
                logger.debug('Ticket have been drawed already, thus return drawed details')
                
                logger.debug('drawed_prize_sequence_indexes=%s', drawed_prize_sequence_indexes)
                
                return create_api_message(status_code=StatusCode.OK, 
                                            won_prize                       = lucky_draw_ticket.drawed_details.get('won_prize'),
                                            drawed_prize_sequence_indexes   = drawed_prize_sequence_indexes,
                                            selected_index                  = lucky_draw_ticket.drawed_details.get('selected_index'),
                                            )
        else:
            return create_api_message('Invalid lucky draw ticket', status_code=StatusCode.BAD_REQUEST)
            
            
    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)
    
@lucky_draw_api_bp.route('/<ticket_key>/read', methods=['GET'])
#@user_auth_token_required
def read_lucky_draw(ticket_key):
    
    db_client = create_db_client(caller_info="read_lucky_draw")
    
    try:
        with db_client.context():
            
            lucky_draw_ticket   = LuckyDrawTicket.get_by_ticket_key(ticket_key)
            logger.debug('lucky_draw_ticket=%s', lucky_draw_ticket)
        
        
        if lucky_draw_ticket:
            
                
            return create_api_message(status_code=StatusCode.OK, 
                                        won_prize                       = lucky_draw_ticket.drawed_details.get('won_prize'),
                                        drawed_prize_sequence_indexes   = lucky_draw_ticket.drawed_details.get('drawed_prize_sequence_indexes'),
                                        selected_index                  = lucky_draw_ticket.drawed_details.get('selected_index'),
                                        )
        else:
            return create_api_message('Invalid lucky draw ticket', status_code=StatusCode.BAD_REQUEST)
            
            
    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)    
    
@lucky_draw_api_bp.route('/<ticket_key>/remove', methods=['DELETE'])
@user_auth_token_required
def remove_lucky_draw_ticket(reference_code, ticket_key):

    db_client = create_db_client(caller_info="remove_lucky_draw_ticket")
    try:
        with db_client.context():
            
            lucky_draw_ticket   = LuckyDrawTicket.get_by_ticket_key(ticket_key)
            logger.debug('lucky_draw_ticket=%s', lucky_draw_ticket)
        
        
        if lucky_draw_ticket:
            with db_client.context():
                lucky_draw_ticket.remove()
        
            return create_api_message(status_code=StatusCode.ACCEPTED)
        
        else:
            return create_api_message('Invalid lucky draw ticket', status_code=StatusCode.BAD_REQUEST)
           
        return create_api_message(status_code=StatusCode.ACCEPTED)    
    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)   
    
@lucky_draw_api_bp.route('/<ticket_key>/grab-prize', methods=['POST'])
@user_auth_token_required
def grab_lucky_draw_prize(reference_code, ticket_key):

    db_client = create_db_client(caller_info="grab_lucky_draw_prize")
    try:
        with db_client.context():
            
            lucky_draw_ticket   = LuckyDrawTicket.get_by_ticket_key(ticket_key)
            logger.debug('lucky_draw_ticket=%s', lucky_draw_ticket)
        
        
        if lucky_draw_ticket:
            
            if lucky_draw_ticket.grabbed:
                return create_api_message('Prize have been grabbed before', status_code=StatusCode.BAD_REQUEST)
            else:
                with db_client.context():
                    customer = lucky_draw_ticket.customer_acct_entity
                    __start_for_lucky_draw_prize(lucky_draw_ticket, customer)
                    
                    
                    customer = Customer.fetch(customer.key_in_str)
                    
                    user_details = customer.registered_user_acct
                    customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                    
                    customer_details_dict['customer_key']               = customer.key_in_str
                    customer_details_dict['is_email_verified']          = user_details.is_email_verified
                    customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                    
                return create_api_message(status_code=StatusCode.OK, customer_rewards=customer_details_dict)
        else:
            return create_api_message('Invalid lucky draw ticket', status_code=StatusCode.BAD_REQUEST)
            
            
    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)        

@model_transactional(desc='grab_lucky_draw_prize')
def __start_for_lucky_draw_prize(lucky_draw_ticket, customer_acct):
    transact_outlet     = lucky_draw_ticket.transact_outlet_entity
    transact_datetime   = datetime.utcnow()
    
    lucky_draw_ticket.grab_the_prize(customer_acct=customer_acct)
    
    lucky_draw_prize_transaction = CustomerTransaction.create_system_transaction(
                                                        customer_acct, 
                                                        system_remarks='Lucky Draw Prize',
                                                        transact_outlet=transact_outlet, 
                                                        transact_datetime=transact_datetime, 
                                                        reward_giveaway_method=program_conf.PROGRAM_REWARD_GIVEAWAY_METHOD_SYSTEM,
                                                        is_sales_transaction = True, 
                                                        allow_to_revert=False, 
                                                        )
    
    program = LuckyDrawRewardProgram(
                drawed_details      = lucky_draw_ticket.drawed_details,
                transact_outlet     = transact_outlet,
                transaction_details = lucky_draw_prize_transaction,
                customer_acct       = customer_acct,
                )
    
    program.give()
    
    
    
@lucky_draw_api_bp.route('/<ticket_key>/view', methods=['GET'])
def view_lucky_draw(ticket_key):

    db_client = create_db_client(caller_info="lucky_draw")
    try:
        with db_client.context():
            lucky_draw_details = LuckyDrawTicket.get_by_ticket_key(ticket_key)
            if lucky_draw_details:
                lucky_draw_details.patch_prize_image_base_url()
                lucky_draw_details = lucky_draw_details.to_dict()
        
        if lucky_draw_details:
            return jsonify(lucky_draw_details)
        else:
            return create_api_message('Fail to draw', status_code=StatusCode.BAD_REQUEST)    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)    


@lucky_draw_api_bp.route('/merchant/<merchant_key>/create-draw-ticket', methods=['GET'])
def create_merchant_lucky_draw_ticket(merchant_key):

    db_client = create_db_client(caller_info="create_merchant_lucky_draw_ticket")
    try:
        logger.debug('merchant_key=%s', merchant_key)
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(merchant_key)
            draw_ticket = LuckyDrawTicket.create(merchant_acct)
        
        logger.debug('draw_ticket=%s', draw_ticket)
        
        if draw_ticket:
            return create_api_message(status_code=StatusCode.OK, draw_ticket=draw_ticket.to_dict())
        else:
            return create_api_message('Failed to create lucky draw ticket', status_code=StatusCode.BAD_REQUEST)    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)

@lucky_draw_api_bp.route('/create-draw-link', methods=['GET'])
@auth_token_required
def create_lucky_draw_link():
    
    acct_id   = request.headers.get('x-acct-id')
    db_client = create_db_client(caller_info="create_lucky_draw_ticket")
    try:
        logger.debug('acct_id=%s', acct_id)
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            draw_ticket = LuckyDrawTicket.create(merchant_acct)
        
        logger.debug('draw_ticket=%s', draw_ticket)
        
        if draw_ticket:
            return create_api_message(status_code=StatusCode.OK, draw_ticket=draw_ticket.to_dict())
        else:
            return create_api_message('Failed to create lucky draw ticket', status_code=StatusCode.BAD_REQUEST)    
    except Exception as err:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message(str(err), status_code=StatusCode.BAD_REQUEST)        
        
        
        


'''
Created on 22 Apr 2024

@author: jacklok
'''
import logging

logger = logging.getLogger('lib')

def elapsed_time_trace(debug=False, trace_key=None):
    def wrapper(fn):
        import time
        def elapsed_time_trace_wrapper(*args, **kwargs):
            start = time.time()
            result      = fn(*args, **kwargs)
            end = time.time()
            elapsed_time = end - start
            trace_name      = trace_key or fn.func_name
            first_argument  = args[0] if args else None
            logger.info('==================== Start Elapsed Time Trace %s(%s) ===========================', trace_name, first_argument)
            logger.info('elapsed time=%s', ("%.2gs" % (elapsed_time)))
            logger.info('================================================================================')
            return result

        return elapsed_time_trace_wrapper
    return wrapper



'''
Created on 14 Dec 2023

@author: jacklok
'''
from flask import Blueprint, request
import logging
from trexlib.utils.log_util import get_tracelog
from trexmodel.utils.model.model_util import create_db_client
from trexapi.utils.api_helpers import StatusCode, create_api_message
from trexmodel.models.datastore.customer_models import Customer
from trexapi.decorators.api_decorators import customer_key_required
from flask_babel import gettext
from datetime import datetime
from trexmodel.models.datastore.redemption_catalogue_models import RedemptionCatalogue
    
from trexapi.utils.redemption_catalogue_helper import giveaway_redeem_catalogue_item

redemption_api_bp = Blueprint('redemption_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/redemption')

logger = logging.getLogger('api')


@redemption_api_bp.route('/list-redemption-catalogue', methods=['GET'])
@customer_key_required
def list_redemption_catalogue(customer_key):
    logger.info('customer_key=%s', customer_key)
    db_client = create_db_client(caller_info="list_redemption_catalogue")
    filtered_catalogues_list = []
    
    with db_client.context():
        customer   = Customer.fetch(customer_key)
        
        merchant_acct = customer.registered_merchant_acct
        
        published_redemption_catalogue_configuration = merchant_acct.published_redemption_catalogue_configuration
        if published_redemption_catalogue_configuration:
            catalogues_list = published_redemption_catalogue_configuration.get('catalogues')
            
            if len(catalogues_list)>0:
                for catalogue in catalogues_list:
                    if __check_is_still_active(catalogue):
                        if __check_is_eligible(customer, catalogue.get('exclusivity')):
                            catalogue['items'] = __resolve_catalogue_items_details(catalogue.get('items'), merchant_acct.published_voucher_configuration.get('vouchers'))
                            filtered_catalogues_list.append(catalogue)
    
    logger.debug('filtered_catalogues_list=%s', filtered_catalogues_list)
                
    return create_api_message(status_code=StatusCode.OK, catalogues_list=filtered_catalogues_list) 

@redemption_api_bp.route('/list-redemption-catalogue/reward-type/<reward_type>', methods=['GET'])
@customer_key_required
def list_redemption_catalogue_by_reward_type(customer_key, reward_type):
    logger.info('customer_key=%s', customer_key)
    logger.info('reward_type=%s', reward_type)
    db_client = create_db_client(caller_info="list_redemption_catalogue_by_reward_type")
    filtered_catalogues_list = []
    
    with db_client.context():
        customer   = Customer.fetch(customer_key)
        
        merchant_acct = customer.registered_merchant_acct
        
        published_redemption_catalogue_configuration = merchant_acct.published_redemption_catalogue_configuration
        if published_redemption_catalogue_configuration:
            catalogues_list = published_redemption_catalogue_configuration.get('catalogues')
            
            if len(catalogues_list)>0:
                for catalogue in catalogues_list:
                    if catalogue.get('redeem_reward_format')==reward_type:
                        if __check_is_still_active(catalogue):
                            if __check_is_eligible(customer, catalogue.get('exclusivity')):
                                catalogue['items'] = __resolve_catalogue_items_details(catalogue.get('items'), merchant_acct.published_voucher_configuration.get('vouchers'))
                                filtered_catalogues_list.append(catalogue)
    
    logger.debug('filtered_catalogues_list=%s', filtered_catalogues_list)
                
    return create_api_message(status_code=StatusCode.OK, catalogues_list=filtered_catalogues_list)    

def __resolve_catalogue_items_details(catalogue_items_list, vouchers_list):
    resolved_items_list = []
    voucher_dict = __convert_vouchers_list_to_dict(vouchers_list)
    for item in catalogue_items_list:
        voucher = voucher_dict.get(item.get('voucher_key'))
        if voucher:
            resolved_items_list.append(
                    {
                        'voucher_key'           : item.get('voucher_key'),
                        'amount'                : item.get('voucher_amount'),
                        'label'                 : voucher.get('label'),
                        'image_url'             : voucher.get('image_url'),
                        'terms_and_conditions'  : voucher.get('terms_and_conditions'),
                        'redeem_reward_amount'  : item.get('redeem_reward_amount'),
                    }
                
                )
        else:
            resolved_items_list.append(item)
        
    return resolved_items_list
        

def __convert_vouchers_list_to_dict(vouchers_list):
    voucher_dict = {}
    for voucher in vouchers_list:
        voucher_dict[voucher.get('voucher_key')] = voucher
    
    return voucher_dict

def __check_is_still_active(catalogue):
    today = datetime.utcnow().date()
    start_date  = datetime.strptime(catalogue.get('start_date'), '%d-%m-%Y').date()
    end_date    = datetime.strptime(catalogue.get('end_date'), '%d-%m-%Y').date()
    
    if today>=start_date and today<=end_date:
        logger.info('catalogue is still valid')
        return True
    else:
        logger.info('catalogue is expired')
        return False

def __check_is_eligible(customer, catalogue_exclusivity):
    
    catalogue_tags_list                 = catalogue_exclusivity.get('tags',[])
    catalogue_memberships_list          = catalogue_exclusivity.get('memberships',[])
    catalogue_tier_memberships_list     = catalogue_exclusivity.get('tier_memberships',[])
    
    is_eligiable = False
    
    if len(catalogue_tags_list)==0 and len(catalogue_memberships_list)==0 and len(catalogue_tier_memberships_list)==0:
        logger.info('catalogue no exclusivity')
        is_eligiable = True
    else:
        if len(catalogue_tags_list)>0:
            logger.info('Going to check member tagging')
            tags_list = customer.tags_list
            for tag in catalogue_tags_list:
                if tag in tags_list:
                    is_eligiable = True
                    logger.info('Found member tagging match with catalogue exclusivity')
                    break
        
        
        if is_eligiable==False and len(catalogue_memberships_list)>0:
            logger.info('Going to check member basic membership')
            memberships_list_list = customer.memberships_list
            for membership in catalogue_memberships_list:
                if membership in memberships_list_list:
                    is_eligiable = True
                    logger.info('Found basic membership match with catalogue exclusivity')
                    break
        
        if is_eligiable==False and len(catalogue_tier_memberships_list)>0:
            logger.info('Going to check member tier membership')
            tier_membership_key = customer.tier_membership_key
            logger.info('customer tier_membership_key=%s', tier_membership_key)
            logger.info('catalogue tier_membership=%s', catalogue_tier_memberships_list)
            if tier_membership_key in catalogue_tier_memberships_list:
                is_eligiable = True
                logger.info('Found tier membership match with catalogue exclusivity')
    
    
    return is_eligiable

@redemption_api_bp.route('/redeem-redemption-catalogue-item', methods=['POST'])
@customer_key_required
def redeem_redemption_catalogue_item(customer_key):
    
    redeem_data_in_json         = request.get_json()
    redemption_catalogue_key    = redeem_data_in_json.get('redemption_catalogue_key')
    voucher_key                 = redeem_data_in_json.get('voucher_key')
    
    customer            = None
    redeem_item_details = None
    
    
    logger.info('customer_key=%s', customer_key)
    logger.info('voucher_key=%s', voucher_key)
    
    
    try:
        db_client = create_db_client(caller_info="redeem_redemption_catalogue_item")
        with db_client.context():
            redemption_catalogue    = RedemptionCatalogue.fetch(redemption_catalogue_key)
            if redemption_catalogue:
                redeem_item_details     = next(filter(lambda x: x['voucher_key'] == voucher_key, redemption_catalogue.catalogue_settings.get('items')), None)
                
                if redeem_item_details:
                    customer = Customer.fetch(customer_key)
                    if customer:
                        reward_summary_before     = customer.reward_summary
                        logger.debug('reward_summary_before=%s', reward_summary_before)
                        giveaway_redeem_catalogue_item(customer, redemption_catalogue, voucher_key)
                        reward_summary_after      = customer.reward_summary
                        logger.debug('reward_summary_after=%s', reward_summary_after)
                        
                    
    except:
        logger.error('Failed to redeem due to %s', get_tracelog())
        return create_api_message(gettext('Failed to redeem'), status_code=StatusCode.BAD_REQUEST)            
            
    if redeem_item_details is None:
        return create_api_message(gettext('Invalid redeem item'), status_code=StatusCode.BAD_REQUEST)    
    elif customer is None:    
        return create_api_message(gettext('Invalid customer profile'), status_code=StatusCode.BAD_REQUEST)
    
    with db_client.context():
        customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
            
    #customer_details_json =  jsonify(customer_details_dict)
    
    return create_api_message(customer_latest_reward=customer_details_dict, status_code=StatusCode.OK)
         


'''
Created on 14 Jul 2021

@author: jacklok
'''

from wtforms import StringField, DecimalField, validators
from trexadmin.forms.base_forms import ValidationBaseForm
from trexadmin.libs.wtforms import validators as custom_validator
from trexadmin.libs.wtforms.fields import OptionalDateTimeField, CurrencyField, JSONField
from flask_babel import gettext
from trexapi.forms.sales_api_forms import SalesTransactionForm

class GiveRewardTransactionForm(SalesTransactionForm):
    sales_amount                    = DecimalField('Sales Amount',[
                                                validators.InputRequired(message="Sales Amount is required"),
                                            ])
    
    tax_amount                      = DecimalField('Tax Amount',[
                                                validators.Optional()
                                            ])  
    
    invoice_id                      = StringField('Invoice No',[
                                                validators.Optional(),
                                                validators.Length(max=30, message="Invoice No length must not more than 30 characters")
                                            ])
    
    remarks                         = StringField('Remarks',[
                                                validators.Optional(),
                                                validators.Length(max=300, message="Remarks length must not more than 300 characters")
                                            ])
    
    transact_datetime               = OptionalDateTimeField('Transact Datetime', format='%d-%m-%Y %H:%M:%S') 
    
class RedeemRewardTransactionForm(ValidationBaseForm):
    reward_format                   = StringField('Reward format',[
                                                validators.InputRequired(message="Reward format is required"),
                                            ])
    
    reward_amount                   = DecimalField('Reward Amount',[
                                                validators.InputRequired(message="Reward amount is required"),
                                            ])  
    
    invoice_id                      = StringField('Invoice No',[
                                                validators.Optional(),
                                                validators.Length(max=30, message="Invoice No length must not more than 30 characters")
                                            ])
    
    remarks                         = StringField('Remarks',[
                                                validators.Optional(),
                                                validators.Length(max=300, message="Remarks length must not more than 300 characters")
                                            ])
    
    redeem_datetime                 = OptionalDateTimeField('Transact Datetime', format='%d-%m-%Y %H:%M:%S')    
     
    
    
    
class VoucherRedeemForm(ValidationBaseForm):
    redeem_code                     = StringField('Voucher Code',[
                                                validators.InputRequired(message="Voucher Code is required"),
                                            ])
    invoice_id                      = StringField('Invoice No',[
                                                validators.Optional(),
                                                validators.Length(max=30, message="Invoice No length must not more than 30 characters")
                                            ])
    
    remarks                         = StringField('Remarks',[
                                                validators.Optional(),
                                                validators.Length(max=300, message="Remarks length must not more than 300 characters")
                                            ])
    
    redeem_datetime               = OptionalDateTimeField('Redeem Datetime', format='%d-%m-%Y %H:%M:%S')  
    
class VoucherRemoveForm(ValidationBaseForm):
    redeem_code                     = StringField('Voucher Code',[
                                                validators.InputRequired(message="Voucher Code is required"),
                                            ])
    remarks                         = StringField('Remarks',[
                                                validators.Optional(),
                                                validators.Length(max=300, message="Remarks length must not more than 300 characters")
                                            ])
    
    remove_datetime               = OptionalDateTimeField('Remove Datetime', format='%d-%m-%Y %H:%M:%S')      
            
    
class PrepaidTopupForm(ValidationBaseForm):
    prepaid_program_key                  = StringField('Prepaid program Key', [
                                            validators.DataRequired(message=gettext("Prepaid program key is required")),
                                            ]
                                            )
    
    topup_amount                        = DecimalField('Topup Amount',[
                                                validators.InputRequired(message="Topup amount is required"),
                                            ])
    
    invoice_id                          = StringField('Invoice Id', [
                                                validators.Length(max=50)
                                            ]
                                            )
    
    remarks                             = StringField('Remarks', [
                                                validators.Length(max=500)
                                            ]
                                            )
    
class PrepaidRedeemForm(ValidationBaseForm):
    redeem_amount                        = DecimalField('Redeem Amount',[
                                                validators.InputRequired(message="Redeem amount is required"),
                                            ])
    
    invoice_id                          = StringField('Invoice Id', [
                                                validators.Length(max=50)
                                            ]
                                            )
    
    remarks                             = StringField('Remarks', [
                                                validators.Length(max=500)
                                            ]
                                            )  
    
class PointRedeemForm(ValidationBaseForm):
    redeem_amount                        = DecimalField('Redeem Amount',[
                                                validators.InputRequired(message="Redeem amount is required"),
                                            ])
    
    invoice_id                          = StringField('Invoice Id', [
                                                validators.Length(max=50)
                                            ]
                                            )
    
    remarks                             = StringField('Remarks', [
                                                validators.Length(max=500)
                                            ]
                                            )        
    
    


'''
Created on 29 Nov 2023

@author: jacklok
'''
from flask import Blueprint, request
import logging
from trexmodel.utils.model.model_util import create_db_client
from datetime import datetime
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.user_models import User
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet
from trexapi.utils.api_helpers import StatusCode,\
    create_api_message
from trexapi.decorators.api_decorators import user_auth_token_required
from flask.json import jsonify
from trexmodel.models.datastore.rating_models import OutletRating,\
    OutletRatingResult, MerchantRatingResult

rating_api_bp = Blueprint('rating_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/rating')

logger = logging.getLogger('api')

@rating_api_bp.route('/outlet/<outlet_key>/review', methods=['POST'])
@user_auth_token_required
def give_outlet_rating(reference_code, outlet_key):
    logger.debug('---give_outlet_rating---')
    
    rating_in_json              = request.get_json()
    service_rating              = rating_in_json.get('service_rating')
    ambience_rating             = rating_in_json.get('ambience_rating')
    food_rating                 = rating_in_json.get('food_rating')
    value_rating                = rating_in_json.get('value_rating')
    
    
    db_client = create_db_client(caller_info="give_outlet_rating")
    
    logger.debug('give_outlet_rating: user account by reference code=%s', reference_code)
    logger.debug('give_outlet_rating: outlet_key=%s', outlet_key)
    
    with db_client.context():
        user_acct       = User.get_by_reference_code(reference_code)
        outlet          = Outlet.fetch(outlet_key)
        merchant_acct   = outlet.merchant_acct_entity
    
    if user_acct and outlet:
        with db_client.context():
            OutletRating.create(user_acct, outlet, 
                                    service_rating  = service_rating, 
                                    ambience_rating = ambience_rating, 
                                    food_rating     = food_rating, 
                                    value_rating    = value_rating,
                                    )
            MerchantRatingResult.update(merchant_acct)
        
        return create_api_message(status_code=StatusCode.OK)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)

@rating_api_bp.route('/outlet/<outlet_key>/review', methods=['GET'])
def read_outlet_rating(outlet_key):   
    logger.debug('give_outlet_rating: outlet_key=%s', outlet_key)
    
    db_client = create_db_client(caller_info="read_outlet_rating")
    
    with db_client.context():
        outlet      = Outlet.fetch(outlet_key) 
        outlet_rating_result = OutletRatingResult.get_by_outlet(outlet)
    
    
    if outlet_rating_result:
        return jsonify(outlet_rating_result.rating_result)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)
    
@rating_api_bp.route('/merchant/<merchant_acct_key>/review', methods=['GET'])
def read_merchant_rating(merchant_acct_key):   
    logger.debug('read_merchant_rating: merchant_acct_key=%s', merchant_acct_key)
    
    db_client = create_db_client(caller_info="read_merchant_rating")
    
    with db_client.context():
        merchant_acct           = MerchantAcct.fetch(merchant_acct_key) 
        merchant_rating_result  = MerchantRatingResult.get_by_merchant_acct(merchant_acct)
    
    
    if merchant_rating_result:
        return jsonify(merchant_rating_result.rating_result)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    
    
@rating_api_bp.route('/outlet/<outlet_key>/update', methods=['POST'])
def update_outlet_rating(outlet_key):   
    logger.debug('update_outlet_rating: outlet_key=%s', outlet_key)
    update_rating_in_json       = request.get_json()
    updated_datetime_from       = update_rating_in_json.get('updated_datetime_from')
    
    db_client = create_db_client(caller_info="update_outlet_rating")
    
    if is_not_empty(updated_datetime_from):
        updated_datetime_from = datetime.strptime(updated_datetime_from, '%d-%m-%Y %H:%M')
    
    with db_client.context():
        outlet      = Outlet.fetch(outlet_key) 
        OutletRatingResult.update(outlet, updated_datetime_from)
    
    
    return create_api_message(triggered_datetime=datetime.now(), status_code=StatusCode.OK)

@rating_api_bp.route('/merchant/<merchant_acct_key>/update', methods=['POST'])
def update_merchant_rating(merchant_acct_key):   
    logger.debug('update_merchant_rating: merchant_acct_key=%s', merchant_acct_key)
    update_rating_in_json       = request.get_json()
    updated_datetime_from       = update_rating_in_json.get('updated_datetime_from')
    
    db_client = create_db_client(caller_info="update_merchant_rating")
    
    if is_not_empty(updated_datetime_from):
        updated_datetime_from = datetime.strptime(updated_datetime_from, '%d-%m-%Y %H:%M')
    
    with db_client.context():
        merchant_acct      = MerchantAcct.fetch(merchant_acct_key) 
        MerchantRatingResult.update(merchant_acct, updated_datetime_from)
    
    
    return create_api_message(triggered_datetime=datetime.now(), status_code=StatusCode.OK)    
    



'''
Created on 30 Jun 2021

@author: jacklok
'''

from flask import Blueprint, request
import flask 
from flask_restful import Resource, abort
import logging
from trexlib.utils.log_util import get_tracelog
from flask_httpauth import HTTPBasicAuth
from trexconf import conf
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.merchant_models import MerchantUser,\
    MerchantAcct, Outlet
import hashlib
from trexlib.utils.string_util import is_not_empty
from flask.json import jsonify
from trexapi import conf as api_conf
from trexapi.decorators.api_decorators import auth_token_required
from trexmodel.models.datastore.loyalty_models import LoyaltyDeviceSetting
from trexmodel.models.datastore.pos_models import POSSetting
from trexapi.controllers.api_routes import APIBaseResource
from trexmodel.models import merchant_helpers
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership, CustomerTierMembership
from trexmodel.models.datastore.rating_models import MerchantRatingResult
from trexmodel.models.datastore.membership_models import MerchantMembership
from trexapi.utils.api_helpers import StatusCode, create_api_message
from trexapi.libs.api_decorator import elapsed_time_trace
from trexmodel.models.datastore.user_models import User

#logger = logging.getLogger('api')
logger = logging.getLogger('debug')

auth = HTTPBasicAuth()
#auth = HTTPBasicAuthWrapper()


merchant_api_bp = Blueprint('merchant_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/merchant')

merchant_api = Api(merchant_api_bp)



@merchant_api_bp.route('/ping', methods=['GET'])
def ping():
    return 'Pong', 200

@auth.verify_password
def verify_user_auth(username, password):
    if not (username and password):
        return False
    
    db_client   = create_db_client(caller_info="verify_user_auth")
    valid_auth  = False
    
    logger.debug('username=%s', username)
    logger.debug('password=%s', password)
    
    with db_client.context():
        merchant_user = MerchantUser.get_by_username(username)
        
        logger.debug('merchant_user=%s', merchant_user)
        
        if merchant_user:
            
            md5_hashed_password = hashlib.md5(password.encode('utf-8')).hexdigest()
            
            logger.debug('verify_user_auth: md5_hashed_password=%s', md5_hashed_password)
            
            if merchant_user.is_valid_password(md5_hashed_password):
                valid_auth = True
            else:
                logger.warn('Invalid merchant password')
        else:
            logger.warn('Invalid merchant username=%s', username)    
        
    return valid_auth

@merchant_api_bp.route('/<merchant_acct_code>/details', methods=['GET'])
def read_merchant_acct(merchant_acct_code):
    
    merchant_acct           = None
    if is_not_empty(merchant_acct_code):
        db_client = create_db_client(caller_info="read_merchant_acct")
        with db_client.context():
            merchant_acct   = MerchantAcct.get_by_account_code(merchant_acct_code)
            
            logger.debug('website=%s', merchant_acct.website)
            merchant_info = merchant_helpers.construct_merchant_acct_info(merchant_acct) 
            
            merchant_rating_result  = MerchantRatingResult.get_by_merchant_acct(merchant_acct)
            
            
            
            if merchant_rating_result:
                logger.debug('rating_result=%s', merchant_rating_result.rating_result)
                merchant_info['rating_reviews'] = merchant_rating_result.rating_result
            
                
        logger.debug('merchant_info=%s', merchant_info)
            
        return create_api_message(status_code=StatusCode.OK,
                                               **merchant_info
                                               )
    else:
        return create_api_message("Missing merchant account key", status_code=StatusCode.BAD_REQUEST)
    
@merchant_api_bp.route('/<merchant_acct_code>/referred-code/<referrer_code>/read-details', methods=['GET'])
def read_customer_brief_and_merchant_acct_and_referred_code(merchant_acct_code, referrer_code):
    
    reference_code = request.args.get('reference_code') or request.form.get('reference_code')
    
    logger.debug('merchant_acct_code=%s', merchant_acct_code)
    logger.debug('referrer_code=%s', referrer_code)
    logger.debug('reference_code=%s', reference_code)
    
    db_client = create_db_client(caller_info="read_customer_brief_and_merchant_acct_from_referred_code")
    customer_acct = None
    
    with db_client.context():
        merchant_acct   = MerchantAcct.get_by_account_code(merchant_acct_code)
        
    if is_not_empty(reference_code):
        with db_client.context():
            customer_acct   = Customer.get_by_reference_code(reference_code, merchant_acct)
            
    
    if is_not_empty(merchant_acct_code):
        
        with db_client.context():
            merchant_info   = merchant_helpers.construct_merchant_acct_info(merchant_acct) 
            referrer_user   = User.get_by_referral_code(referrer_code)
            
            if merchant_info and referrer_user:
                
                if merchant_info.get('referral_program_settings'):
                    
                    refer_a_friend_url      = '{base_url}/referral/program/merchant-code/{merchant_code}/referrer-code/{referrer_code}/join'
                    refer_a_friend_message  = 'Hi, \n\n{referee_promote_desc}. Please join {brand_name} via below link:\n\n{refer_a_friend_url}'
                    referee_promote_desc    = merchant_info.get('referral_program_settings').get('referee_promote_desc')
                        
                    refer_a_friend_url = refer_a_friend_url.format(
                                                base_url        = conf.REFER_BASE_URL,
                                                merchant_code   = merchant_acct.account_code,
                                                referrer_code   = referrer_code,
                                                )
                    
                    refer_a_friend_deep_link = conf.REFER_A_FRIEND_DEEP_LINK.format(
                                                merchant_acct_code  = merchant_acct.account_code,
                                                referrer_code       = referrer_code,
                                                )
                    
                    
                    refer_a_friend_message = refer_a_friend_message.format(
                                                referee_promote_desc    = referee_promote_desc,
                                                brand_name              = merchant_acct.brand_name,
                                                refer_a_friend_url      = refer_a_friend_url,
                                                )
                    
                    merchant_info['referral_program_settings']['refer_a_friend_url']        = refer_a_friend_url
                    merchant_info['referral_program_settings']['refer_a_friend_message']    = refer_a_friend_message
                    merchant_info['referral_program_settings']['refer_a_friend_deep_link']  = refer_a_friend_deep_link
                    
                merchant_rating_result  = MerchantRatingResult.get_by_merchant_acct(merchant_acct)
                
                
                
                if merchant_rating_result:
                    logger.debug('rating_result=%s', merchant_rating_result.rating_result)
                    merchant_info['rating_reviews'] = merchant_rating_result.rating_result
        
        if referrer_user is None or merchant_info is None:  
            
            return create_api_message(status_code=StatusCode.BAD_REQUEST)
        else:   
                
            logger.debug('merchant_info=%s', merchant_info)
            
            data = {
                        'merchant_info' : merchant_info,
                        'referrer_name' : referrer_user.name,
                        }
            
            if customer_acct is not None:
                data['customer_brief'] = {
                                        'name'              : customer_acct.name,
                                        'reference_code'    : reference_code,
                                        }
                
            return create_api_message(status_code=StatusCode.OK,
                                                   **data
                                                   )
    else:
        return create_api_message("Missing merchant account key", status_code=StatusCode.BAD_REQUEST)    
    
@merchant_api_bp.route('/<merchant_acct_code>/joined-brand-details/reference-code/<reference_code>', methods=['GET'])
@elapsed_time_trace(trace_key='read_joined_brand_details')
def read_joined_brand_details(merchant_acct_code, reference_code):
    
    merchant_acct           = None
    
    if is_not_empty(merchant_acct_code) and is_not_empty(reference_code):    
        db_client = create_db_client(caller_info="read_merchant_acct")
        with db_client.context():
            merchant_acct           = MerchantAcct.get_by_account_code(merchant_acct_code)
            
            customer                = Customer.get_by_reference_code(reference_code, merchant_acct)
            if customer:
                user_acct               = customer.registered_user_acct
                merchant_info           = merchant_helpers.construct_merchant_acct_info(merchant_acct) 
                merchant_rating_result  = MerchantRatingResult.get_by_merchant_acct(merchant_acct)
                
                if merchant_info.get('referral_program_settings'):
                    
                    refer_a_friend_url      = '{base_url}/referral/program/merchant-acct-code/{merchant_acct_code}/referrer-code/{referrer_code}/join'
                    refer_a_friend_message  = 'Hi, \n\n{referee_promote_desc}. Please join {brand_name} via below link:\n\n{refer_a_friend_url}'
                    referee_promote_desc    = merchant_info.get('referral_program_settings').get('referee_promote_desc')
                        
                    refer_a_friend_url = refer_a_friend_url.format(
                                                base_url            = conf.REFER_BASE_URL,
                                                merchant_acct_code  = merchant_acct.account_code,
                                                referrer_code       = user_acct.referral_code,
                                                )
                    
                    refer_a_friend_deep_link = conf.REFER_A_FRIEND_DEEP_LINK.format(
                                                merchant_acct_code   = merchant_acct.account_code,
                                                referrer_code   = user_acct.referral_code,
                                                )
                    
                    
                    refer_a_friend_message = refer_a_friend_message.format(
                                                referee_promote_desc    = referee_promote_desc,
                                                brand_name              = merchant_acct.brand_name,
                                                refer_a_friend_url      = refer_a_friend_url,
                                                )
                    
                    merchant_info['referral_program_settings']['refer_a_friend_url']        = refer_a_friend_url
                    merchant_info['referral_program_settings']['refer_a_friend_message']    = refer_a_friend_message
                    merchant_info['referral_program_settings']['refer_a_friend_deep_link']  = refer_a_friend_deep_link
                
                if merchant_rating_result:
                    logger.debug('rating_result=%s', merchant_rating_result.rating_result)
                    merchant_info['rating_reviews'] = merchant_rating_result.rating_result
                
                user_details                                        = customer.registered_user_acct
                customer_details = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                customer_details['customer_key']               = customer.key_in_str
                customer_details['is_email_verified']          = user_details.is_email_verified
                customer_details['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                
                customer_basic_memberships_list = _list_customer_basic_memberships(customer)
                if customer_basic_memberships_list:
                    customer_details['basic_memberships'] = customer_basic_memberships_list
                
                tier_membership_data = _get_tier_membership(customer)
                if tier_membership_data:
                    customer_details['tier_membership'] = tier_membership_data
                
        logger.debug('merchant_info=%s', merchant_info)
        
        if customer:    
            return create_api_message(status_code=StatusCode.OK,
                                               merchant_info    = merchant_info,
                                               customer_details = customer_details,
                                               )
        else:
            return create_api_message(status_code=StatusCode.BAD_REQUEST)                                       
    else:
        return create_api_message("Missing merchant account key", status_code=StatusCode.BAD_REQUEST)    

def _list_customer_basic_memberships(customer):
    customer_membership_final_list = []
    
    if is_not_empty(customer.memberships_list):
        customer_memberships_list = CustomerMembership.list_all_by_customer(customer)
        if is_not_empty(customer_memberships_list):
            merchant_acct = customer.registered_merchant_acct
            merchant_memberships_list = MerchantMembership.list_by_merchant_acct(merchant_acct)
            
            for cm in customer_memberships_list:
                for mm in merchant_memberships_list:
                    logger.debug('cm=%s', cm)
                    if mm.key_in_str == cm.merchant_membership_key:
                        membership_data = {
                                                        'key'                   : mm.key_in_str,
                                                        'label'                 : mm.label,
                                                        'entitled_date'         : cm.entitled_date.strftime('%d-%m-%Y'),
                                                        'expiry_date'           : cm.expiry_date.strftime('%d-%m-%Y'),
                                                        'card_image'            : mm.membership_card_image,
                                                        'desc'                  : mm.desc if is_not_empty(mm.desc) else '',
                                                        'terms_and_conditions'  : mm.terms_and_conditions if is_not_empty(mm.terms_and_conditions) else '',
                                                        'is_tier'               : False,
                                                        }
                        
                        if cm.renewed_date is not None:
                            membership_data['renewed_date'] = cm.renewed_date.strftime('%d-%m-%Y'),
                        
                        customer_membership_final_list.append(membership_data)
                        break
    return customer_membership_final_list

def _get_tier_membership(customer):
            
    if is_not_empty(customer.tier_membership):
        merchant_tier_membership = customer.tier_membership_entity
        customer_tier_membership = CustomerTierMembership.get_by_customer_and_merchant_tier_membership(customer, merchant_tier_membership)
        
        membership_data = {
                        'key'                   : customer_tier_membership.merchant_tier_membership_key,
                        'label'                 : merchant_tier_membership.label,
                        'entitled_date'         : customer_tier_membership.entitled_date.strftime('%d-%m-%Y'),
                        'expiry_date'           : customer_tier_membership.expiry_date.strftime('%d-%m-%Y'),
                        'card_image'            : merchant_tier_membership.membership_card_image,
                        'desc'                  : merchant_tier_membership.desc if is_not_empty(merchant_tier_membership.desc) else '',
                        'terms_and_conditions'  : merchant_tier_membership.terms_and_conditions if is_not_empty(merchant_tier_membership.terms_and_conditions) else '',
                        'is_tier'               : True,
                        }
        
        return membership_data

@merchant_api_bp.route('/<merchant_key>/customer-details/reference-code/<reference_code>', methods=['GET'])
def read_merchant_customer_details(merchant_key, reference_code):
    
    merchant_acct           = None
    customer_details_dict   = {} 
    
    if is_not_empty(merchant_key) and is_not_empty(reference_code):
        db_client = create_db_client(caller_info="read_merchant_customer_details")
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(merchant_key)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct) 
        
            if customer:
            
                user_details = customer.registered_user_acct
                customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                customer_details_dict['customer_key']               = customer.key_in_str
                customer_details_dict['is_email_verified']          = user_details.is_email_verified
                customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                
                customer_basic_memberships_list = _list_customer_basic_memberships(customer)
                if customer_basic_memberships_list:
                    customer_details_dict['basic_memberships'] = customer_basic_memberships_list
                
                tier_membership_data = _get_tier_membership(customer)
                if tier_membership_data:
                    customer_details_dict['tier_membership'] = tier_membership_data
                
                '''
                if 'entitled_voucher_summary' in customer_details_dict:
                    customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
                    del customer_details_dict['entitled_voucher_summary']
                '''
        return jsonify(customer_details_dict)
            
    
    else:
        return create_api_message("Missing merchant account key or user reference code", status_code=StatusCode.BAD_REQUEST)


class AccountActivatedAPIResource(APIBaseResource):  
    
    @auth.login_required
    def post(self):
        username    = auth.current_user()
        
        acct_id     = request.headers.get('x-acct-id')
        api_key     = request.headers.get('x-api-key')
        outlet_key  = request.headers.get('x-outlet-key')
        
        logger.debug('username=%s', username)
        logger.debug('acct_id=%s', acct_id)    
        logger.debug('api_key=%s', api_key)
        logger.debug('outlet_key=%s', outlet_key)
        
        is_valid            = False
        is_authorized       = False
        output_json = {}
        
        if is_not_empty(acct_id) and is_not_empty(api_key):
            db_client = create_db_client(caller_info="AccountActivatedAPIResource.post")
            with db_client.context():
                merchant_acct = MerchantAcct.fetch(acct_id)
                
                #logger.debug('merchant_acct=%s', merchant_acct)
                
                if merchant_acct:
                    logger.debug('merchant_acct api key=%s', merchant_acct.api_key)
                    
                    if api_key == merchant_acct.api_key:
                        merchant_user               = MerchantUser.get_by_username(username)
                        
                        is_authorized = self.is_outlet_authorized(merchant_user, outlet_key)
                        if is_authorized:
                            (expiry_datetime, token)    = self.generate_token(acct_id, username)
                            logger.debug('outlet is_authorized=%s', is_authorized)
                            logger.debug('expiry_datetime=%s', expiry_datetime)
                            
                            output_json =  {
                                            'auth_token'        : token,
                                            'expires_in'        : int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE) * 60,
                                            #'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %h:%M:$S'),
                                            'granted_outlet'    : merchant_user.granted_outlet_details_list,
                                            'username'          : merchant_user.username,
                                            'name'              : merchant_user.name,
                                            'is_admin'          : merchant_user.is_admin,
                                            'granted_access'    : merchant_user.permission.get('granted_access'),
                                            'gravatar_url'      : merchant_user.gravatar_url,
                                            }
                
                            logger.debug('output_json=%s', output_json)
                            
                            is_valid = True
                    
            if is_valid:
                return output_json
            
            elif is_authorized==False:
                abort(400, msg=["User is not authorized due to outlet is not granted"])
                    
        abort(400, msg=["Failed to authenticate"])
        
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        return True 
    
class AuthenticateAPIResource(APIBaseResource):  
    
    @auth.login_required
    def post(self):
        username    = auth.current_user()
        
        acct_id     = request.headers.get('x-acct-id')
        api_key     = request.headers.get('x-api-key')
        outlet_key  = request.headers.get('x-outlet-key')
        
        logger.debug('acct_id=%s', acct_id)    
        logger.debug('api_key=%s', api_key)
        logger.debug('outlet_key=%s', outlet_key)
        
        is_authenticated    = False
        is_authorized       = False
        output_json = {}
        
        if is_not_empty(acct_id) and is_not_empty(api_key):
            db_client = create_db_client(caller_info="AuthenticateAPIResource.post")
            with db_client.context():
                merchant_acct = MerchantAcct.fetch(acct_id)
            
                if merchant_acct:
                    if api_key == merchant_acct.api_key:
                        merchant_user               = MerchantUser.get_by_username(username)
                        
                        is_authorized = self.is_outlet_authorized(merchant_user, outlet_key)
                        logger.debug('outlet is_authorized=%s', is_authorized)
                        
                        if is_authorized:
                            (expiry_datetime, token)    = self.generate_token(acct_id, username)
                            #session['auth_username']    = username
                            #session['acct_id']          = acct_id
                            
                            logger.debug('token=%s', token)
                            logger.debug('expiry_datetime=%s', expiry_datetime)
                            logger.debug('auth_username=%s', username)
                
                            output_json =  {
                                            'auth_token'        : token,
                                            'expires_in'        : int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE) * 60,
                                            #'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %h:%M:$S'),
                                            'granted_outlet'    : merchant_user.granted_outlet_details_list,
                                            'username'          : merchant_user.username,
                                            'name'              : merchant_user.name,
                                            'is_admin'          : merchant_user.is_admin,
                                            'granted_access'    : merchant_user.permission.get('granted_access'),
                                            'gravatar_url'      : merchant_user.gravatar_url,
                                            }
                
                            logger.debug('output_json=%s', output_json)
                            
                            is_authenticated = True
                    
            if is_authenticated:
                return output_json
            elif is_authorized==False:
                abort(400, msg=["User is not authorized due to outlet is not granted"])
                    
        abort(400, msg=["Failed to authenticate"])
        
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        return True    

class ProgramDeviceAuthenticate(AccountActivatedAPIResource):
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        if merchant_user.is_admin:
            logger.debug('it is merchant admin user, thus it is authorized for all outlet')
            return True
        else:
            data_in_json        = request.get_json()
            activation_code     = data_in_json.get('activation_code')
            device_setting      = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
            access_outlet       = Outlet.fetch(outlet_key)
            assigned_outlet_key = device_setting.assigned_outlet_key
            
            logger.debug('device assigned outlet name=%s', device_setting.assigned_outlet_entity.name)
            logger.debug('access outlet name=%s', access_outlet.name)
            
            if outlet_key == device_setting.assigned_outlet_key:
                logger.debug('login outlet is same as device setting assigned outlet')
                if outlet_key in merchant_user.granted_outlet:
                    logger.debug('login outlet is authorized by login merchant user granted outlet')
                    if assigned_outlet_key in merchant_user.granted_outlet:
                        logger.debug('device assigned outlet is authorized by login merchant user granted outlet')
                        return True
                    else:
                        logger.debug('device setting assigned outlet is not granted outlet')
                else:
                    logger.debug('login outlet is not granted outlet')    
            else:
                logger.debug('login outlet is not same as device setting assigned outlet')
                        
        return False
            
class POSDeviceAuthenticate(AccountActivatedAPIResource):
    def is_outlet_authorized(self, merchant_user, outlet_key):
        if merchant_user.is_admin:
            return True
        else:
            data_in_json        = request.get_json()
            activation_code     = data_in_json.get('activation_code')
            device_setting      = POSSetting.get_by_activation_code(activation_code)
            access_outlet       = Outlet.fetch(outlet_key)
            assigned_outlet_key = device_setting.assigned_outlet_key
            
            logger.debug('device assigned outlet name=%s', device_setting.assigned_outlet_entity.name)
            logger.debug('access outlet name=%s', access_outlet.name)
            
            if outlet_key == device_setting.assigned_outlet_key:
                logger.debug('login outlet is same as device setting assigned outlet')
                if outlet_key in merchant_user.granted_outlet:
                    logger.debug('login outlet is authorized by login merchant user granted outlet')
                    if assigned_outlet_key in merchant_user.granted_outlet:
                        logger.debug('device assigned outlet is authorized by login merchant user granted outlet')
                        return True
                    else:
                        logger.debug('device setting assigned outlet is not granted outlet')
                else:
                    logger.debug('login outlet is not granted outlet')    
            else:
                logger.debug('login outlet is not same as device setting assigned outlet')
        
        return False            

class SecureAPIResource(AuthenticateAPIResource):  
    
    def __init__(self):
        super(SecureAPIResource, self).__init__()  
    
    
class CheckAuthTokenResource(SecureAPIResource):
    
    @auth_token_required
    def get(self):
        return 'Ping'             
    
     
merchant_api.add_resource(AuthenticateAPIResource,       '/auth')
merchant_api.add_resource(ProgramDeviceAuthenticate,     '/program-auth')
merchant_api.add_resource(POSDeviceAuthenticate,         '/pos-auth')      
    
        
        
        
        


'''
Created on 30 Jan 2024

@author: jacklok
'''
from firebase_admin import credentials, messaging

def create_prepaid_push_notification(title_data=None, speech=None, message_data=None, device_token=None):
    message = messaging.Message(
        notification=messaging.Notification(
            title   = title_data,
            body    = message_data,
            
        ),
        data={
            'speech': speech,
        },
        token=device_token,
    )
    
    messaging.send(message)



'''
Created on 1 Jul 2021

@author: jacklok
'''
from functools import wraps
from flask import request, session, abort
from trexlib.utils.string_util import is_not_empty
from trexlib.utils.crypto_util import decrypt_json
import logging
from datetime import datetime
from trexlib.utils.log_util import get_tracelog
from trexadmin.libs.http import create_rest_message, StatusCode
from trexmodel.models.datastore.loyalty_models import LoyaltyDeviceSetting
from trexmodel.models.datastore.pos_models import POSSetting
from trexmodel.utils.model.model_util import create_db_client

logger = logging.getLogger('decorator')
#logger = logging.getLogger('debug')

def test_session_expired(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        return ("Authenticated token is expired", 401)
    
    return decorated_function

def auth_token_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        auth_token  = request.headers.get('x-auth-token')
        acct_id     = request.headers.get('x-acct-id')
            
        logger.debug('acct_id=%s', acct_id)
        logger.debug('auth_token=%s', auth_token)
        
        if is_not_empty(auth_token):
            try:
                auth_details_json = decrypt_json(auth_token)
            except:
                logger.error('Authenticated token is not valid')
                return ("Authenticated token is not valid", 401)
            
            logger.debug('auth_details_json=%s', auth_details_json)
            
            if auth_details_json:
                expiry_datetime     = auth_details_json.get('expiry_datetime')
                acct_id_from_token  = auth_details_json.get('acct_id')
                
                logger.debug('acct_id from decrypted token=%s', acct_id_from_token)
                logger.debug('expiry_datetime from decrypted token=%s', expiry_datetime)
                
                if is_not_empty(expiry_datetime) and is_not_empty(acct_id_from_token) and acct_id==acct_id_from_token:
                    expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                    logger.debug('expiry_datetime=%s', expiry_datetime)
                    
                    now             = datetime.now()
                    if now < expiry_datetime: 
                        logger.debug('auth token is still valid')
                        return f(*args, **kwargs)
                    else:
                        logger.debug('auth token is not logger valid')
                        
                        return create_rest_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED,)
                else:
                    return create_rest_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED,)
        
        return create_rest_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED,)

    return decorated_function

def show_request_info(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        logger.debug('request url=%s', request.url)
        return f(*args, **kwargs)
    
    return decorated_function
        

def user_auth_token_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_auth_token     = request.headers.get('x-auth-token')
        user_reference_code = request.headers.get('x-reference-code')
        
        logger.debug('user_auth_token=%s', user_auth_token)
        
        logger.debug('args=%s', args)
        logger.debug('kwargs=%s', kwargs)
        
        if is_not_empty(user_auth_token):
            logger.debug('user_auth_token is not empty, going to decrypt it')
            try:
                auth_details_json = decrypt_json(user_auth_token)
                logger.debug('auth_details_json=%s', auth_details_json)
                
            except:
                logger.error('Failed due to %s', get_tracelog())
                return create_rest_message('Authenticated token is not valid', status_code=StatusCode.UNAUTHORIZED,)
            
            
            
            if auth_details_json:
                expiry_datetime                 = auth_details_json.get('expiry_datetime')
                user_reference_code_from_token  = auth_details_json.get('reference_code')
                
                if is_not_empty(expiry_datetime) and is_not_empty(user_reference_code_from_token) and user_reference_code_from_token == user_reference_code:
                    expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                    logger.debug('expiry_datetime=%s', expiry_datetime)
                    
                    now             = datetime.now()
                    if now < expiry_datetime: 
                        logger.debug('auth token is still valid')
                        logger.debug('new args=%s', (user_reference_code, *args))
                        logger.debug('new kwargs=%s', kwargs)
                        
                        return f(*(user_reference_code, *args), **kwargs)
                        
                    else:
                        logger.debug('auth token is not logger valid')
                        
                        #return ("Authenticated token is expired", 401)
                        return create_rest_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED,)
            else:
                #return ("Authenticated token is invalid", 401)    
                return create_rest_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED,)
        
        #return ("Authenticated token is required", 401)
        return create_rest_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED,)

    return decorated_function

def device_is_activated(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        activation_code     = request.headers.get('x-activation-code')
        device_id           = request.headers.get('x-device-id')
        device_type         = request.headers.get('x-device-type')
            
        logger.debug('activation_code=%s', activation_code)
        logger.debug('device_id=%s', device_id)
        logger.debug('device_type=%s', device_type)
        
        if is_not_empty(activation_code) and is_not_empty(device_id):
            device_settings = None
            try:
                db_client = create_db_client(caller_info="device_is_activated")
                if 'loyalty' == device_type:
                    with db_client.context():
                        device_settings = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
                elif 'pos' == device_type:
                    with db_client.context():
                        device_settings = POSSetting.get_by_activation_code(activation_code)
                
                if device_settings is not None:
                    if device_settings.testing:
                        logger.info('Activation device id is meant for testing thus, it can be by pass')
                    else:
                        if device_settings.device_id != device_id:
                            logger.info('Access device id is not same as activation device id')
                            return create_rest_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                    
                    return f(*args, **kwargs)
                
                return create_rest_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                
            except:
                logger.error('failed due to %s ', get_tracelog())
                #return ("Failed to check device authorization", 401)
                return create_rest_message('Failed to check device authorization', status_code=StatusCode.BAD_REQUEST)
            
            

    return decorated_function

def device_activated_is_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        activation_code     = request.headers.get('x-activation-code')
        device_id           = request.headers.get('x-device-id')
        device_type         = request.headers.get('x-device-type')
            
        logger.debug('activation_code=%s', activation_code)
        logger.debug('device_id=%s', device_id)
        logger.debug('device_type=%s', device_type)
        
        if is_not_empty(activation_code) and is_not_empty(device_id):
            device_settings = None
            try:
                db_client = create_db_client(caller_info="device_is_activated")
                if 'loyalty' == device_type:
                    with db_client.context():
                        device_settings = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
                elif 'pos' == device_type:
                    with db_client.context():
                        device_settings = POSSetting.get_by_activation_code(activation_code)
                
                if device_settings is not None:
                    if device_settings.testing:
                        logger.info('Activation device id is meant for testing thus, it can be by pass')
                    else:
                        if device_settings.device_id != device_id:
                            logger.info('Access device id is not same as activation device id')
                            return create_rest_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                    
                    return f(*(activation_code, *args), *args, **kwargs)
                
                return create_rest_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                
            except:
                logger.error('failed due to %s ', get_tracelog())
                #return ("Failed to check device authorization", 401)
                return create_rest_message('Failed to check device is activated', status_code=StatusCode.BAD_REQUEST)
            
        else:  
            return create_rest_message('Missing activation code, or device type or device id', status_code=StatusCode.BAD_REQUEST)  

    return decorated_function

def outlet_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        outlet_key = request.headers.get('x-outlet-key')
            
        logger.debug('outlet_key=%s', outlet_key)
        
        if is_not_empty(outlet_key):
            logger.debug('Going to execute')
            return f(*args, **kwargs)
            
        
        #return ("Outlet Key is required", 401)
        return create_rest_message('Missing outlet key', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function

def merchant_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        acct_id = request.headers.get('x-acct-id')
            
        logger.debug('acct_id=%s', acct_id)
        
        if is_not_empty(acct_id):
            logger.debug('Going to execute')
            return f(*args, **kwargs)
            
        
        #return ("Merchant account id is required", 401)
        return create_rest_message('Missing merchant account id', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function

def customer_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        customer_key = request.headers.get('x-customer-key')
            
        logger.debug('customer_key=%s', customer_key)
        
        if is_not_empty(customer_key):
            logger.debug('Going to execute')
            return f(customer_key, **kwargs)
            
        
        #return ("Outlet Key is required", 401)
        return create_rest_message('Missing customer key', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function


'''
Created on 19 Jul 2021

@author: jacklok
'''

from flask import Blueprint, request, session 
import logging
from trexlib.utils.log_util import get_tracelog
from trexmodel.utils.model.model_util import create_db_client
from datetime import datetime, timedelta
from trexapi.decorators.api_decorators import auth_token_required,\
    user_auth_token_required, device_is_activated
from trexlib.utils.string_util import is_not_empty, is_empty
from trexmodel.models.datastore.customer_models import Customer
from trexapi.utils.api_helpers import create_api_message, StatusCode
from trexmodel.models.datastore.merchant_models import Outlet,\
    MerchantUser, MerchantAcct
from trexapi.forms.reward_api_forms import VoucherRedeemForm, VoucherRemoveForm
from werkzeug.datastructures import ImmutableMultiDict
import json
from trexapi.utils.api_helpers import get_logged_in_api_username
from trexmodel.models.datastore.voucher_models import MerchantVoucher
from trexmodel.models.datastore.reward_models import CustomerEntitledVoucher

from trexmodel.models.datastore.redeem_models import CustomerRedemption
from trexanalytics.bigquery_upstream_data_config import create_merchant_customer_redemption_upstream_for_merchant
from trexmodel import program_conf
from trexmodel.models.datastore.user_models import User
from flask.json import jsonify
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.reward_model_helpers import check_redeem_voucher_is_valid
from flask_babel import gettext
from trexlib.libs.flask_wtf.request_wrapper import request_json, request_headers

logger = logging.getLogger('api')


voucher_api_bp = Blueprint('voucher_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/vouchers')

logger = logging.getLogger('api')

@voucher_api_bp.route('/ping', methods=['GET'])
def ping():
    return create_api_message('Pong', status_code=StatusCode.OK)

@voucher_api_bp.route('/voucher/<redeem_code>/details', methods=['POST','GET'])
@auth_token_required
@device_is_activated
def read_voucher(redeem_code):
    if is_not_empty(redeem_code):
        voucher_details = None
        db_client = create_db_client(caller_info="read_voucher")
        with db_client.context():
            customer_voucher    = CustomerEntitledVoucher.get_by_redeem_code(redeem_code)
            if customer_voucher:
                merchant_voucher    = MerchantVoucher.fetch(customer_voucher.entitled_voucher_key)
                voucher_conf        = merchant_voucher.configuration
                if merchant_voucher.configuration:
                    if isinstance(merchant_voucher.configuration, str):
                        voucher_conf = json.loads(merchant_voucher.configuration)
                else:
                    voucher_conf = {}
                voucher_details = {
                                    'key'                   : merchant_voucher.key_in_str,
                                    'label'                 : merchant_voucher.label,
                                    'desc'                  : merchant_voucher.desc,
                                    'terms_and_conditions'  : merchant_voucher.terms_and_conditions,
                                    'configuration'         : voucher_conf,
                                    'image_url'             : merchant_voucher.image_public_url,
                                    'redeem_info_list'      : [
                                                                {
                                                                    'effective_date'    : customer_voucher.effective_date.strftime('%d-%m-%Y'),
                                                                    'expiry_date'       : customer_voucher.expiry_date.strftime('%d-%m-%Y'),
                                                                    'redeem_code'       : redeem_code,
                                                                    'is_redeem'         : customer_voucher.is_used,
                                                                    'is_revert'         : customer_voucher.is_reverted,
                                                                }
                                                            ],
                                    }
                
        
        if voucher_details is None:
            return create_api_message('Invalid voucher code', status_code=StatusCode.BAD_REQUEST)
        else:
            #return create_api_message(voucher_details=voucher_details, status_code=StatusCode.OK)
            
            logging.debug('voucher_details=%s', voucher_details)
            
            return jsonify(voucher_details)
    else:
        return create_api_message('Voucher code is required', status_code=StatusCode.BAD_REQUEST) 
    
@voucher_api_bp.route('/voucher/redeem', methods=['post'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
def redeem_voucher(redeem_voucher_data_in_json, request_headers):
    
    #redeem_voucher_data_in_json   = request.get_json()
    
    redeem_voucher_form = VoucherRedeemForm(ImmutableMultiDict(redeem_voucher_data_in_json))
    
    if redeem_voucher_form.validate():
    
        redeem_code         = redeem_voucher_data_in_json.get('redeem_code')
        invoice_id          = redeem_voucher_form.invoice_id.data
        remarks             = redeem_voucher_form.remarks.data
        redeemed_datetime   = redeem_voucher_form.redeem_datetime.data
        merchant_acct       = None
        
        logger.debug('redeem_code=%s', redeem_code)
        logger.debug('redeemed_datetime=%s', redeemed_datetime)
        logger.debug('redeem_code=%s', redeem_code)
        
        
        if redeem_code:
            redeem_code_list = redeem_code.split(',')
            
            
            db_client = create_db_client(caller_info="redeem_voucher")
            with db_client.context():
                redeemed_by_outlet      = Outlet.fetch(request_headers.get('x-outlet-key'))
                merchant_acct           = MerchantAcct.fetch(request_headers.get('x-acct-id'))
            
            redeem_datetime             = redeem_voucher_form.redeem_datetime.data
            merchant_username           = get_logged_in_api_username()
            
            
            if redeem_datetime:
                redeem_datetime_in_gmt      = redeem_datetime - timedelta(hours=merchant_acct.gmt_hour)
                now                         = datetime.utcnow()
                if redeem_datetime_in_gmt > now:
                    return create_api_message('Redeem datetime cannot be future', status_code=StatusCode.BAD_REQUEST)
            
            
            
            to_redeem_voucher_keys_list                             = []
            customer                                                = None
            customer_vouchers_list                                  = []
            
            with db_client.context():
                merchant_username       = get_logged_in_api_username()
                redeemed_by             = MerchantUser.get_by_username(merchant_username)
                
                for redeem_code in redeem_code_list:
                    customer_voucher    = CustomerEntitledVoucher.get_by_redeem_code(redeem_code)
                    customer_vouchers_list.append(customer_voucher)
                    to_redeem_voucher_keys_list.append(customer_voucher.key_in_str)
                    if customer is None:
                        customer = customer_voucher.entitled_customer_entity
                    else:
                        checking_customer = customer_voucher.entitled_customer_entity
                        if checking_customer.key_in_str!=customer.key_in_str:
                            return create_api_message(gettext('Voucher must from same customer'), status_code=StatusCode.BAD_REQUEST)
                            
                
            try:
                with db_client.context():
                    check_redeem_voucher_is_valid(customer, customer_vouchers_list, redeem_datetime=redeem_datetime)
            except Exception as error:
                logger.error('Failed due to %s', get_tracelog())
                return create_api_message(str(error), status_code=StatusCode.BAD_REQUEST)
            
            @model_transactional(desc='redeem_voucher')
            def __start_transaction_for_redeem_voucher():
                
                customer_redemption = CustomerRedemption.create(customer, program_conf.REWARD_FORMAT_VOUCHER , 
                                                          redeemed_outlet               = redeemed_by_outlet,
                                                          redeemed_amount               = 1,            
                                                          redeemed_voucher_keys_list    = to_redeem_voucher_keys_list, 
                                                          redeemed_by                   = redeemed_by, 
                                                          redeemed_datetime             = redeemed_datetime,
                                                          invoice_id                    = invoice_id,
                                                          remarks                       = remarks,
                                                          )
                return customer_redemption
            
            if to_redeem_voucher_keys_list:
                with db_client.context():
                    if to_redeem_voucher_keys_list:
                        
                        customer_redemption = __start_transaction_for_redeem_voucher()
                        
                        logger.debug('customer_redemption=%s', customer_redemption)
                        
        
                    if customer_redemption:
                        create_merchant_customer_redemption_upstream_for_merchant(customer_redemption, streamed_datetime=redeemed_datetime)
                
                if customer_redemption:        
                    return create_api_message(transaction_id = customer_redemption.transaction_id, status_code=StatusCode.OK)
                else:
                    return create_api_message("Failed to redeem voucher", status_code=StatusCode.BAD_REQUEST)
            else:
                return create_api_message('Voucher redeem code is required', status_code=StatusCode.BAD_REQUEST)
                
    else:
        logger.warn('redeem voucher data input is invalid')
        error_message = redeem_voucher_form.create_rest_return_error_message()
            
        return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)
        
@voucher_api_bp.route('/voucher/<redeem_code>/remove', methods=['DELETE'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
def remove_voucher(remove_voucher_data_in_json, request_headers, redeem_code):
    
    #remove_voucher_data_in_json   = request.get_json()
    
    remove_voucher_form = VoucherRemoveForm(ImmutableMultiDict(remove_voucher_data_in_json))
    if remove_voucher_form.validate():
    
        #redeem_code         = remove_voucher_data_in_json.get('redeem_code')
        remarks             = remove_voucher_form.remarks.data
        remove_datetime     = remove_voucher_form.remove_datetime.data
        #merchant_acct       = None
        
        logger.debug('redeem_code=%s', redeem_code)
        logger.debug('remove_datetime=%s', remove_datetime)
        
        
        if redeem_code:
            #redeem_code_list = redeem_code_list.split(',')
            redeem_code_list = redeem_code.split(',')
            
            db_client = create_db_client(caller_info="redeem_voucher")
            '''
            with db_client.context():
                merchant_acct           = MerchantAcct.fetch(request.headers.get('x-acct-id'))
            '''
            remove_datetime_in_gmt      = remove_voucher_form.remove_datetime.data
            
            to_remove_voucher_keys_list                         = []
            found_not_valid_redeem_code_list                    = []
            removed_by                                          = None    
            
            with db_client.context():
                merchant_acct           = MerchantAcct.fetch(request_headers.get('x-acct-id'))
                merchant_username       = get_logged_in_api_username()
                removed_by              = MerchantUser.get_by_username(merchant_username)
                
                
                for redeem_code in redeem_code_list:
                    customer_voucher    = CustomerEntitledVoucher.get_by_redeem_code(redeem_code)
                    if customer_voucher:
                        to_remove_voucher_keys_list.append(customer_voucher.key_in_str)
                    else:
                        found_not_valid_redeem_code_list.append(redeem_code)
            
            if found_not_valid_redeem_code_list:
                return create_api_message("Voucher ({redeem_codes_list}) is not valid".format(redeem_codes_list=",".join(found_not_valid_redeem_code_list)), 
                                           status_code=StatusCode.BAD_REQUEST)            
            else:
                if to_remove_voucher_keys_list:
                    with db_client.context():
                        if to_remove_voucher_keys_list:
                            for voucher_key in to_remove_voucher_keys_list:
                                customer_voucher = CustomerEntitledVoucher.fetch(voucher_key)
                                customer_voucher.remove(removed_by, removed_datetime=remove_datetime_in_gmt)
                            
                    return create_api_message(status_code=StatusCode.OK)
                    
                else:
                    return create_api_message('Voucher redeem code is required', status_code=StatusCode.BAD_REQUEST)
    
    else:
        logger.warn('redeem voucher data input is invalid')
        error_message = remove_voucher_form.create_rest_return_error_message()
            
        return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)        
             

@voucher_api_bp.route('/entitled/<reference_code>', methods=['GET'])
def list_entitled_voucher(reference_code):
    
    logger.debug('list_entitled_voucher: going to list entitled voucher by reference code=%s', reference_code)
    
    if is_not_empty(reference_code):
        db_client       = create_db_client(caller_info="list_entitled_voucher")
        user_acct       = None
        voucher_list    = []
        
        
        with db_client.context():
            user_acct = User.get_by_reference_code(reference_code)
        
            if user_acct:
                result = CustomerEntitledVoucher.list_all_by_user_acct(user_acct)
                if result:
                    for r in result:
                        voucher_list.append(r.to_dict())    
            
            
        
                                       
        return {
                'vouchers': voucher_list,
            }
            
    else:
        logger.warn('reset_password_post: email is invalid')
        return create_api_message(status_code=StatusCode.BAD_REQUEST)       
    
@voucher_api_bp.route('/voucher/<redeem_code>/remove-by-user', methods=['DELETE'])
@user_auth_token_required
@request_headers
def remove_user_voucher(reference_code, request_headers, redeem_code):
    logger.info('reference_code=%s', reference_code)
    logger.info('redeem_code=%s', redeem_code)
    
    if is_not_empty(redeem_code):
        reference_code = request_headers.get('x-reference-code')
        logger.debug('reference_code=%s', reference_code)
        db_client = create_db_client(caller_info="remove_user_voucher")
        with db_client.context():
            customer_voucher    = CustomerEntitledVoucher.get_by_redeem_code(redeem_code)
            
            if customer_voucher: 
                
                customer_voucher.remove()  
                customer = customer_voucher.entitled_customer_acct
                customer.update_after_removed_voucher(customer_voucher)
            else:
                logger.info('voucher is not found')    
               
        if customer_voucher and customer_voucher.is_removed:        
        
            return create_api_message(status_code=StatusCode.OK)
        else:
            logger.info('Invalid voucher redeem code')
            return create_api_message('Invalid voucher redeem code', status_code=StatusCode.BAD_REQUEST)
    else:
        return create_api_message('Missing voucher redeem code', status_code=StatusCode.BAD_REQUEST)
